from dataclasses import dataclass, field
from chardet import detect as chdetect
from pandas import DataFrame
from proxysiphon.proxychimp import Guts
import proxysiphon.lgm as lgm


def read_ncdc(filepath_or_buffer, encoding=None):
    """Read NOAA NCDC txt file

    Parameters
    ----------
    filepath_or_buffer
    encoding : str or None, optional
        File encoding. Default is None which attempts to guess the encoding with
        `chardet.detect`.

    Returns
    -------
    out : NcdcRecord
    """
    with open(filepath_or_buffer, 'rb') as fl:
        flbytes = fl.read()
    if encoding is None:
        encoding = chdetect(flbytes)['encoding']
    g = Guts(flbytes.decode(encoding))
    return g.to_ncdcrecord()


def read_lgm(filepath_or_buffer, encoding=None):
    """Read NOAA NCDC txt file for LGM proxies

    Parameters
    ----------
    filepath_or_buffer
    encoding : str or None, optional
        File encoding. Default is None which attempts to guess the encoding with
        `chardet.detect`.

    Returns
    -------
    out : NcdcRecord
    """
    out = read_ncdc(filepath_or_buffer, encoding=None)
    return LgmRecord(**out.__dict__)


def read_petm(filepath_or_buffer, encoding=None):
    """Read NOAA NCDC txt file for PETM proxies

    Parameters
    ----------
    filepath_or_buffer
    encoding : str or None, optional
        File encoding. Default is None which attempts to guess the encoding with
        `chardet.detect`.

    Returns
    -------
    out : NcdcRecord
    """
    out = read_ncdc(filepath_or_buffer, encoding=None)
    return PetmRecord(**out.__dict__)


@dataclass
class SiteInformation:
    """Proxy site information"""
    site_name: str = None
    location: str = None
    country: str = None
    northernmost_latitude: float = None
    southernmost_latitude: float = None
    easternmost_longitude: float = None
    westernmost_longitude: float = None
    elevation: float = None


@dataclass
class DataCollection:
    """Proxy site data collection information"""
    collection_name: str = None
    first_year: float = None
    last_year: float = None
    time_unit: str = None
    core_length: str = None
    notes: str = None
    collection_year: int = None


@dataclass
class VariableInfo:
    """Proxy site Data Variable information"""
    what: str
    material: str
    error: str
    units: str
    seasonality: str
    archive: str
    detail: str
    method: str
    datatype: str


@dataclass
class ChronologyInformation:
    """Proxy site chronology information"""
    df: DataFrame = field(default_factory=DataFrame)


@dataclass
class Data:
    """Proxy site data variables"""
    df: DataFrame = field(default_factory=DataFrame)


@dataclass
class Publication:
    """Proxy site publication"""
    authors: str = None
    published_date_or_year: int = None
    published_title: str = None
    journal_name: str = None
    volume: str = None
    edition: str = None
    issue: str = None
    pages: str = None
    report_number: str = None
    doi: str = None
    online_resource: str = None
    full_citation: str = None
    abstract: str = None

    def to_citationstr(self):
        """Citation str of publication"""
        if self.full_citation is not None:
            return str(self.full_citation)

        out = '{authors} ({year}): {title}.'.format(authors=self.authors,
                                                    year=self.published_date_or_year,
                                                    title=self.published_title)
        # This is a bit lazy...
        if self.journal_name is not None:
            out += ' {},'.format(self.journal_name)
        if self.edition is not None:
            out += ' {},'.format(self.edition)
        if self.volume is not None:
            out += ' {},'.format(self.volume)
        if self.issue is not None:
            out += ' {},'.format(self.issue)
        if self.pages is not None:
            out += ' {},'.format(self.pages)
        if self.report_number is not None:
            out += ' {},'.format(self.report_number)
        if self.doi is not None:
            out += ' doi:{}'.format(self.doi)
        if self.online_resource is not None:
            out += ' {}'.format(self.online_resource)

        return out


@dataclass
class NcdcRecord:
    """Proxy site NCDC record"""
    chronology_information: ChronologyInformation = None
    data: Data = None
    data_collection: DataCollection = None
    description: str = None
    original_source_url: str = None
    publication: list = field(default_factory=list)
    site_information: SiteInformation = None
    variables: dict = field(default_factory=dict)


class LgmRecord(lgm.QcPlotMixin, lgm.RedateMixin, lgm.NetcdfMixin, NcdcRecord):
    """Proxy site LGM NCDC record"""
    pass


class PetmRecord(LgmRecord):
    """PETM proxy site NCDC record"""
    pass



from tempfile import NamedTemporaryFile
import pytest
import pandas as pd

from proxysiphon import proxychimp


datapayload = ['#------------------------',
           '# Contribution_Date',
           '#',
           '#------------------------',
           '# Title',
           '#',
           '#------------------------',
           '# Chronology_Information',
           '#',
           '# Labcode\tdepth_top\tdepth_bottom\tmat_dated\t14C_date\t14C_1s_err\tdelta_R\tdelta_R_1s_err\tother_date\tother_1s_err\tother_type\t',
           '# 152757	5	6	G. ruber or mixed planktonic	475	30	188	73	-999	-999	-999	',
           '#',
           '#------------------------',
           '# Data',
           '# Data lines follow (have no #)',
           '# Missing Value: -999',
           'depth\tage\tbacon',
           '1\t2\t3',
           '4\t5\t6']


@pytest.fixture(scope='module')
def chron_nodeltaR_nodata_guts():
    payload = ['#------------------------',
               '# Contribution_Date',
               '#',
               '#------------------------',
               '# Title',
               '#',
               '#------------------------',
               '# Chronology_Information',
               '#',
               '# Labcode\tdepth_top\tdepth_bottom\tmat_dated\t14C_date\t14C_1s_err\tdelta_R\tdelta_R_1s_err\tother_date\tother_1s_err\tother_type\t',
               '# 152757	5	6	G. ruber or mixed planktonic	-999	-999	-999	-999	-999	-999	-999	',
               '#',
               '#------------------------',
               '# Data',
               '# Data lines follow (have no #)',
               '# Missing Value: -999',
               'depth\tage\tbacon']
    with NamedTemporaryFile('wb') as tf:
        tf.write('\n'.join(payload).encode('utf-8'))
        tf.flush()
        g = proxychimp.Guts(tf.name)
    return g


@pytest.fixture(scope='module')
def chron_guts():
    payload = datapayload
    with NamedTemporaryFile('wb') as tf:
        tf.write('\n'.join(payload).encode('utf-8'))
        tf.flush()
        g = proxychimp.Guts(tf.name)
    return g


@pytest.fixture(scope='module')
def dumb_guts():
    payload = ['#------------------------',
               '# NOTE: Please cite original publication, online resource and date accessed when using this data.',
               '# If there is no publication information, please cite Investigator,', '#',
               '# Description/Documentation lines begin with #', '# Data lines have no #', '#',
               '# Online_Resource: http://www.ncdc.noaa.gov/paleo/study/',
               '# Online_Resource: http://www1.ncdc.noaa.gov/pub/data/paleo/', '#',
               '# Original_Source_URL: https://www.ncdc.noaa.gov/paleo-search/study/2622',
               '#------------------------',
               '# Contribution_Date',
               '#',
               '#------------------------',
               '# Title',
               '#',
               '#------------------------',
               '# Data_Collection', '#   Collection_Name: P178-15P',
               '#   First_Year: 39485', '#   Last_Year: -18',
               '#   Time_Unit: cal yr BP', '#   Core_Length: ',
               '#   Notes: mg_red', '#   Collection_Year: 1923',
               '#------------------------',
               '# Site Information', '# Site_Name: P178-15P',
               '# Location: Arabian Sea', '# Country: ',
               '# Northernmost_Latitude: 11.955', '# Southernmost_Latitude: 11.955',
               '# Easternmost_Longitude: 44.3', '# Westernmost_Longitude: 44.3',
               '# Elevation: -869',
               '#------------------------',
               '# Description and Notes',
               '#        Description: d18O sacc from 2003 paper, mg/ca sacc from 2002 paper, alkeno',
               '#------------------------',
               '# Publication',
               '# Authors: Tierney, Jessica E.; Pausata, Francesco S. R.; deMenocal, Peter B.',
               '# Published_Date_or_Year: 2016',
               '# Published_Title: Deglacial Indian monsoon failure and North Atlantic stadials linked by Indian Ocean surface cooling',
               '# Journal_Name: Nature Geoscience', '# Volume: 9',
               '# Edition: ', '# Issue: ', '# Pages: 46-50',
               '# Report Number: ', '# DOI: 10.1038/ngeo2603',
               '# Online_Resource: ', '# Full_Citation: ', '# Abstract:',
               '#------------------------',
               '# Chronology_Information',
               '#',
               '# Labcode\tdepth_top\tdepth_bottom\tmat_dated\t14C_date\t14C_1s_err\tdelta_R\tdelta_R_1s_err\tother_date\tother_1s_err\tother_type\t',
               '#',
               '#---------------------------------------',
               '# Variables',
               '# Data variables follow that are preceded by "##" in columns one and two.',
               '# Variables list, one per line, shortname-tab-longname components (9 components: what, material, error, units, seasonality, archive, detail, method, C or N for Character or Numeric data)',
               '## depth	,,,cm,,,,,',
               '## age	,,,cal yr BP,,,,,',
               '## bacon	,,,index,,,,,',
               '#------------------------',
               '# Data',
               '# Data lines follow (have no #)',
               '# Missing Value: -999',
               'depth\tage\tbacon',
               '1\t2\t3',
               '4\t5\t6']
    with NamedTemporaryFile('wb') as tf:
        tf.write('\n'.join(payload).encode('utf-8'))
        tf.flush()
        g = proxychimp.Guts(tf.name)
    return g


def test_find_values():
    lines = ['apple: 1\n', 'bee: 1:2\n', 'charlie: 1.5\n', 'dingo-2\t\n', 'echo: ']
    assert proxychimp.find_values(lines, 'NotAKey') is None
    assert proxychimp.find_values(lines, 'apple') == '1'
    assert proxychimp.find_values(lines, 'bee') == '1:2'
    assert proxychimp.find_values(lines, 'charlie', fun=float) == 1.5
    assert proxychimp.find_values(lines, 'bee', sep='-', fun=int) is None
    assert proxychimp.find_values(lines, 'dingo', sep='-', fun=int) == 2
    assert proxychimp.find_values(lines, 'echo') is None


def test_str_guts__init_():
    filestr = '\n'.join(datapayload)
    g = proxychimp.Guts(filestr)
    goal_dict = {'Labcode': [152757], 'depth_top': [5], 'depth_bottom': [6],
                 'mat_dated': ['G. ruber or mixed planktonic'],
                 '14C_date': [475],'14C_1s_err': [30],
                 'delta_R': [188], 'delta_R_1s_err': [73],
                 'other_date': [pd.np.nan], 'other_1s_err': [pd.np.nan],
                 'other_type': [pd.np.nan]}
    goal = pd.DataFrame(goal_dict)
    df = g.yank_chron_df()
    for k in goal_dict.keys():
        pd.testing.assert_series_equal(goal[k], df[k])


def test__divide_portions(dumb_guts):
    goal_data = ['depth\tage\tbacon', '1\t2\t3', '4\t5\t6']
    goal_description_ends = ('#------------------------', '# Missing Value: -999')
    goal_beginline = 70
    assert goal_data == dumb_guts.data
    assert goal_description_ends[0] == dumb_guts.description[0]
    assert goal_description_ends[-1] == dumb_guts.description[-1]
    assert goal_beginline == dumb_guts.data_beginline


def test__write_sectionindex(dumb_guts):
    goal_index = {'Chronology_Information': [(55, 59)],
                  'Contribution_Date': [(12, 14)],
                  'Data': [(67, 70)],
                  'Title': [(15, 17)]}
    assert 10 == len(dumb_guts.sectionindex.keys())
    for k, goal in goal_index.items():
        assert goal == dumb_guts.sectionindex[k]


def test_pull_section(dumb_guts):
    goal = ['# Data', '# Data lines follow (have no #)',
            '# Missing Value: -999']
    assert dumb_guts.pull_section('Data')[0] == goal


def test_pull_section_badname(dumb_guts):
    with pytest.raises(KeyError):
        dumb_guts.pull_section('data')

def test_available_sections(dumb_guts):
    goal = ['Contribution_Date', 'Title', 'Data_Collection',
            'Site Information', 'Description and Notes', 'Publication',
            'Chronology_Information', 'Variables', 'Data']
    # Skip first section as it is very long:
    assert goal == dumb_guts.available_sections()[1:]


def test_guess_missingvalues(dumb_guts):
    goal = '-999'
    assert goal == dumb_guts.guess_missingvalues()


def test_yank_data_df(dumb_guts):
    goal_dict = {'depth': [1, 4], 'age': [2, 5], 'bacon': [3, 6]}
    goal = pd.DataFrame(goal_dict)
    df = dumb_guts.yank_data_df()
    for k in goal_dict.keys():
        pd.testing.assert_series_equal(goal[k], df[k])


def test_yank_chron_df(chron_guts):
    goal_dict = {'Labcode': [152757], 'depth_top': [5], 'depth_bottom': [6],
                 'mat_dated': ['G. ruber or mixed planktonic'],
                 '14C_date': [475],'14C_1s_err': [30],
                 'delta_R': [188], 'delta_R_1s_err': [73],
                 'other_date': [pd.np.nan], 'other_1s_err': [pd.np.nan],
                 'other_type': [pd.np.nan]}
    goal = pd.DataFrame(goal_dict)
    df = chron_guts.yank_chron_df()
    for k in goal_dict.keys():
        pd.testing.assert_series_equal(goal[k], df[k])


def test_yank_publication(dumb_guts):
    v1 = dumb_guts.yank_publication()[0]
    assert 'Tierney, Jessica E.; Pausata, Francesco S. R.; deMenocal, Peter B.' == v1['authors']
    assert '10.1038/ngeo2603' == v1['doi']


def test_yank_site_infromation(dumb_guts):
    v = dumb_guts.yank_site_information()
    assert 'P178-15P' == v['site_name']
    assert 'Arabian Sea' == v['location']
    assert v['country'] is None
    assert 11.955 == v['northernmost_latitude']
    assert 11.955 == v['southernmost_latitude']
    assert 44.3 == v['easternmost_longitude']
    assert 44.3 == v['westernmost_longitude']
    assert -869 == v['elevation']


def test_yank_data_collection(dumb_guts):
    v = dumb_guts.yank_data_collection()
    assert 'P178-15P' == v['collection_name']
    assert 39485 == v['first_year']
    assert -18 == v['last_year']
    assert 'cal yr BP' == v['time_unit']
    assert v['core_length'] is None
    assert 'mg_red' == v['notes']
    assert 1923 == v['collection_year']


def test_pull_variables(dumb_guts):
    v = dumb_guts.yank_variables()
    assert 'cm' == v['depth'][3]


def test_yank_description_and_notes(dumb_guts):
    v = dumb_guts.yank_description_and_notes()
    assert 'd18O sacc from 2003 paper, mg/ca sacc from 2002 paper, alkeno' == v


def test_yank_original_source_url(dumb_guts):
    v = dumb_guts.yank_original_source_url()
    assert 'https://www.ncdc.noaa.gov/paleo-search/study/2622' == v


def test_has_chron(dumb_guts, chron_guts):
    assert dumb_guts.has_chron() is False
    assert chron_guts.has_chron() is True


def test_has_deltar(dumb_guts, chron_guts, chron_nodeltaR_nodata_guts):
    assert chron_nodeltaR_nodata_guts.has_deltar() is False
    assert dumb_guts.has_deltar() is False
    assert chron_guts.has_deltar() is True


def test_has_datacolumn(dumb_guts, chron_nodeltaR_nodata_guts):
    assert chron_nodeltaR_nodata_guts.has_datacolumn('boogers') is False
    assert dumb_guts.has_datacolumn('depth') is True



from proxysiphon import records


def test_publication_to_citationstr():
    pub = records.Publication(authors='White, Tom; New, White',
                              published_date_or_year=1986,
                              published_title='Article title',
                              journal_name='Cool Journal',
                              volume=12, issue=3, pages=173,
                              doi='sfdjla/vcxl.3')
    goal = "White, Tom; New, White (1986): Article title. Cool Journal, 12, 3, 173, doi:sfdjla/vcxl.3"
    assert pub.to_citationstr() == goal


import datetime
import logging
from copy import deepcopy

import netCDF4
import numpy as np
import unidecode

from proxysiphon.agemodel import get_deltar_online, fit_agedepthmodel, date_proxy

log = logging.getLogger(__name__)


def _normalize_to_ascii_array(a, dtype='S50'):
    """Normalize sequence of UTF-8 string to np.Array of ASCII"""
    normed = np.array([unidecode.unidecode(str(x)) for x in a], dtype=dtype)
    return normed


class RedateMixin:
    """Mixins to redate LGM proxy records"""

    def swapin_custom_deltar(self, d_r=None, d_std=None):
        """Swap-in custom proxy ΔR mean and standard-deviation values.

        ``d_r`` is put into self.chronology_information.df column "delta_R" and
        ``d_std`` is put into column "delta_R_1s_err". Any existing values in
        these columns are first moved to a new column "*_original", but only if
        the "*_original" columns don't already exist. If "*_original" columns
        already exist, then current values in "delta_R" or "delta_R_1s_err" are
        discarded in the returned proxy record.

        Parameters
        ----------
        d_r : ndarray, scalar or None
            Carbon reservoir (ΔR) mean value to swap-in. Ignored if None.
        d_std : ndarray, scalar or None
            ΔR standard-deviation value to swap-in. Ignored
            if None.

        Returns
        -------
        out : Modified copy of proxy record.
        """
        out = self.copy()

        if d_r is None and d_std is None:
            return out

        # If we're plugging in a new value, we preserve the original but moving it to a new *_original
        # columns.
        if d_r is not None:
            if 'delta_R_original' not in out.chronology_information.df.columns:
                out.chronology_information.df['delta_R_original'] = out.chronology_information.df['delta_R']
            out.chronology_information.df['delta_R'] = d_r
            out.chronology_information.df['delta_R'].replace(to_replace='None', value=np.nan, inplace=True)

        if d_std is not None:
            if 'delta_R_1s_err_original' not in out.chronology_information.df.columns:
                out.chronology_information.df['delta_R_1s_err_original'] = out.chronology_information.df[
                    'delta_R_1s_err']
            out.chronology_information.df['delta_R_1s_err'] = d_std
            out.chronology_information.df['delta_R_1s_err'].replace(to_replace='None', value=np.nan, inplace=True)

        return out

    def copy(self):
        """Return deep copy of self"""
        return deepcopy(self)

    def redate(self, **kwargs):
        """Redate proxy data with (snake)bacon

        Parameters
        ----------
        nsims : int
            Number of age-model simulations to retain.
        **kwargs :
            Passed on to ``self._fit_agemodel``.

        Returns
        -------
        Updated copy of self. The ``snakebacon.AgeDepthModel`` used to redate
        the record is monkeypatched into
        ``self.chronology_information.bacon_agemodel`` in this returned copy.
        """
        rec = self.copy()

        agemodel, mcmc_kwargs = rec._fit_agemodel(**kwargs)
        a_median, a_ensemble = rec._date_sampledepths(agemodel, kwargs.pop('nsims', None))
        rec.chronology_information.bacon_agemodel = agemodel
        rec.data.age_median = a_median
        rec.data.age_ensemble = a_ensemble
        return rec

    def _fit_agemodel(self, **kwargs):
        """Fit snakebacon model to NcdcRecord
        """
        chron_df = self.chronology_information.df.copy()
        data_df = self.data.df.copy()
        myr = 1950 - self.recent_date()
        deltar = self.chronology_information.df['delta_R'].values
        deltar_error = self.chronology_information.df['delta_R_1s_err'].values

        agemodel, _, mcmc_kwargs = fit_agedepthmodel(chron_df, data_df, deltar, deltar_error,
                                                     minyr=myr, mcmc_kws=kwargs)
        return agemodel, mcmc_kwargs

    def _date_sampledepths(self, agemodel, nsims=1000):
        """Date self proxy sample depths given a snakebacon.AgeModel

        Parameters
        ----------
        agemodel : snakebacon.AgeModel
            Age model to date from.
        nsims : scalar or None, optional
            Number of draws to include in returned ensemble. ``None`` returns
            1000 draws.

        Returns
        -------
        p_median : pandas.DataFrame
            Median age per depth.
        p_ensemble :
            Ensemble of ages per depth.
        """
        if nsims is None:
            nsims = 1000

        data_df = self.data.df.copy()
        p_median, p_ensemble = date_proxy(agemodel, data_df, nsims)

        p_median = (p_median[['depth', 'age']].rename(columns={'age': 'age_median'})
                    .set_index('depth'))
        p_ensemble = (p_ensemble[['depth', 'age', 'mciter']].rename(columns={'mciter': 'draw'})
                      .pivot(index='depth', columns='draw', values='age'))
        return p_median, p_ensemble

    def recent_date(self):
        """Get the most recent date from self metadata.

        First try "Collection_Year" in "Data_Collection" section. If can't
        find, then earliest year in publications list.
        """
        out = None
        col_year = self.data_collection.collection_year
        if col_year is not None:
            out = col_year
        else:
            # Return earliest year in publications.
            out = min([int(p.published_date_or_year) for p in self.publication if p.published_date_or_year is not None])
        return out

    def update_deltar(self):
        """Set self.ChronologyInformation.df delta_R and delta_r_1s_error, returning an updated copy

        If these variables are changed/updated, the originals are moved to column names *_original.
        """
        x = self.copy()
        latlon = (float(x.site_information.northernmost_latitude),
                  float(x.site_information.easternmost_longitude))

        chron_df = x.chronology_information.df.copy()

        delta_r_used = None
        delta_r_1s_err_used = None

        n_unique_deltarerror = len(chron_df['delta_R'].dropna().unique())
        n_unique_deltar = len(chron_df['delta_R_1s_err'].dropna().unique())

        if (n_unique_deltarerror > 1) and (n_unique_deltar > 1):
            log.info('Found multi-depth deltar and deltar_error. Using NcdcRecords original values')
        elif n_unique_deltar > 1:
            log.info('Found multi-depth deltar, without deltar_error. Using carbonferret "delta_R_1s_err".')
            delta_r_1s_err_used = get_deltar_online(latlon)[1]
        else:
            delta_r_used, delta_r_1s_err_used = get_deltar_online(latlon)
            log.debug('deltar(deltar_error): {}({})'.format(delta_r_used, delta_r_1s_err_used))

        if delta_r_used is not None:
            chron_df['delta_R_original'] = chron_df['delta_R'].copy()
            chron_df['delta_R'] = delta_r_used

        if delta_r_1s_err_used is not None:
            chron_df['delta_R_1s_err_original'] = chron_df['delta_R_1s_err'].copy()
            chron_df['delta_R_1s_err'] = delta_r_1s_err_used

        x.chronology_information.df = chron_df.copy()
        return x

    def average_duplicate_datadepths(self):
        """Average obs at the same depth in self.data.df, return copy of self
        """
        if not self.data.df['depth'].duplicated().any():
            log.debug('Found no duplicate depths in Data')
            # Return if no duplicates
            return self

        x = self.copy()
        log.debug('Found duplicate depths in Data')
        x.data.df = x.data.df.groupby('depth', as_index=False).mean()
        return x

    def has_chron(self):
        """True if self has populated chronology information"""
        result = False
        try:
            chron = self.chronology_information.df
        except KeyError:
            return result
        if len(chron) > 1:  # Needs to have more than one date.
            result = True
        return result

    def has_data(self):
        """True if self has populated data information"""
        result = False
        try:
            d = self.data.df
        except KeyError:
            return result
        if len(d) > 0:
            result = True
        return result

    def chronology_depth_range(self, chronology='original'):
        """Get (min, max) range of chronology determinant depths

        Parameters
        ----------
        chronology : str, optional
            Which chronology information do we get depth ranges from. Can be
            'original' or 'redate'. 'original' uses the depths in chronology information from
            self.chronology_information.df. 'redate' looks to
            self.chronology_information.bacon_agemodel.mcmcsetup.coredates.depth,
            assuming that the record has been redated with self.readate().

        Returns
        -------
        out : tuple
            Depths range (min, max).

        See Also
        --------
        LgmRecord.slice_datadepths : Cut self.data.df to given depths
        """
        assert chronology in ['original', 'redate']
        if chronology == 'redate':
            depths = np.array(self.chronology_information.bacon_agemodel.mcmcsetup.coredates.depth)
        elif chronology == 'original':
            depths = np.concatenate([self.chronology_information.df['depth_top'],
                                     self.chronology_information.df['depth_bottom']])
        out = (np.nanmin(depths), np.nanmax(depths))
        return out

    def slice_datadepths(self, shallow=None, deep=None):
        """Cut self.data to given depths, return updated copy of self

        The cut is inclusive for `shallow` and `deep` limits. Both `shallow` and
        `deep` must be positive. If no values are given for `shallow` and `deep`,
        the cuts are at the deepest and shallowest samples in
        `self.chronology_information`.

        Note also that it cuts self.data.df, in addition to self.data.age_median
        and self.data.age_ensemble, if available.

        See Also
        --------
        LgmRecord.chronology_depth_range : Get range of chronology determinant depths
        """
        if len(self.data.df) < 1 or 'depth' not in self.data.df.columns:
            return self

        out = self.copy()

        if shallow is None:
            shallow = out.data.df['depth'].min()
        if deep is None:
            deep = out.data.df['depth'].max()

        assert shallow >= 0 and deep >= 0, 'cut depths must be positive'

        out.data.df = out.data.df[(out.data.df['depth'] >= shallow) & (out.data.df['depth'] <= deep)]

        if hasattr(out.data, 'age_ensemble'):
            out.data.age_ensemble = out.data.age_ensemble[
                (out.data.age_ensemble.index >= shallow) & (out.data.age_ensemble.index <= deep)]
        if hasattr(out.data, 'age_median'):
            out.data.age_median = out.data.age_median[
                (out.data.age_median.index >= shallow) & (out.data.age_median.index <= deep)]

        return out


class NetcdfMixin:
    """Mixins to export LGM proxy records to a NetCDF file"""

    @staticmethod
    def _variable_attributes(varname):
        """Get dict of netCDF4 variable attributes for a given NcdcRecord Data column name"""
        proxy_map = {'d13c': ('d13C', 'per mil'),
                     'd18o': ('d18O', 'per mil'),
                     'mgca': ('Mg/Ca', 'per mil'),
                     'percent': ('Percent foraminifera', '%'),
                     'tex86': ('TEX86', 'index'),
                     'uk37': ("UK'37", 'index')}
        # Second part of the column name...
        foraminifera_map = {'bulloides': 'Globigerina bulloides',
                            'crassaformis': 'Globorotalia crassaformis',
                            'dutertrei': 'Neogloboquadrina dutertrei',
                            'inflata': 'Globoconella inflata',
                            'mabahethi': 'Cibicides mabahethi',
                            'marginata': 'Bulimina marginata',
                            'menardii': 'Globorotalia menardii',
                            'obliquiloculata': 'Pulleniatina obliquiloculata',
                            'pachyderma': 'Neogloboquadrina pachyderma sinistral',
                            'pachysin': 'Neogloboquadrina pachyderma sinistral',
                            'pachyderma_d': 'Neogloboquadrina incompta',
                            'peregrina': 'Uvigerina peregrina',
                            'quinqueloba': 'Turborotalita quinqueloba',
                            'ruber': 'Globigerinoides ruber white',
                            'ruber_lato': 'Globigerinoides ruber white',
                            'ruber_pink': 'Globigerinoides ruber pink',
                            'ruber_stricto': 'Globigerinoides ruber white',
                            'sacculifer': 'Trilobatus sacculifer',
                            'truncatulinoides': 'Globorotalia pachytheca',
                            'tumida': 'Globorotalia tumida',
                            'acicula': 'Creseis acicula',
                            }
        varname = varname.lower()
        try:
            if '_' not in varname:
                out = {'long_name': proxy_map[varname][0], 'units': proxy_map[varname][1]}
            else:
                proxy, foram = varname.split('_', 1)
                out = {'long_name': proxy_map[proxy][0], 'units': proxy_map[proxy][1],
                       'foraminifera_type': foraminifera_map[foram]}
        except KeyError:  # Variable name not found.
            out = {}
        return out

    def _attach_site_ncgroup(self, parent):
        """Create the NetCDF4 site group and populate it

        This is run whenever self.to_netcdf() is called.

        Parameters
        ----------
        parent : netcdf4.Group or netcdf4.Dataset
            Object the created site group will use as a parent.

        Returns
        -------
        this_site : netcdf4.Group
            Reference to the created chronology group.
        """
        site_name = self.site_information.site_name.strip()
        # Normalize site group name to ASCII and replace whitespace characters.
        grp_name = unidecode.unidecode(site_name.lower()).replace(' ', '_')

        # Create and populate site group
        this_site = parent.createGroup(grp_name)
        this_site.site_name = str(site_name)
        this_site.comment = str(self.description)
        this_site.latitude = float(self.site_information.northernmost_latitude)
        this_site.longitude = float(self.site_information.easternmost_longitude)
        this_site.elevation = int(self.site_information.elevation)
        try:
            this_site.collection_year = int(self.data_collection.collection_year)
        except TypeError:  # If collection year doesn't exist
            pass
        publication_str = '\n\n'.join([x.to_citationstr() for x in self.publication])
        this_site.references = publication_str
        return this_site

    def _attach_chronology_ncgroup(self, parent):
        chron = parent.createGroup('chronology')
        chron.createDimension('depth_top', None)
        chron.createDimension('str_dim', 50)

        labcode = chron.createVariable('labcode', 'S1', ('depth_top', 'str_dim'))
        labcode.long_name = 'Lab sample code'
        labcode.missing_value = 'NA'
        labcode._Encoding = 'ascii'
        labcode[:] = _normalize_to_ascii_array(self.chronology_information.df['Labcode'].fillna('NA'))

        depth_top = chron.createVariable('depth_top', 'f4', ('depth_top',))
        depth_top.long_name = 'Sample top depth'
        depth_top.axis = 'Z'
        depth_top.positive = 'down'
        depth_top.units = 'cm'
        depth_top.missing_value = np.nan
        depth_top[:] = self.chronology_information.df['depth_top'].values

        depth_bottom = chron.createVariable('depth_bottom', 'f4', ('depth_top',))
        depth_bottom.long_name = 'Sample bottom depth'
        depth_bottom.units = 'cm'
        depth_bottom.positive = 'down'
        depth_bottom.missing_value = np.nan
        depth_bottom[:] = self.chronology_information.df['depth_bottom'].values

        mat_dated = chron.createVariable('mat_dated', 'S1', ('depth_top', 'str_dim'))
        mat_dated.long_name = 'Material dated'
        mat_dated.missing_value = 'NA'
        mat_dated._Encoding = 'ascii'
        mat_dated[:] = _normalize_to_ascii_array(self.chronology_information.df['mat_dated'].fillna('NA'))

        c14_date = chron.createVariable('c14_date', 'f4', ('depth_top',))
        c14_date.long_name = '14C date'
        c14_date.units = 'RC yr BP'
        c14_date.missing_value = np.nan
        c14_date[:] = self.chronology_information.df['14C_date'].values

        c14_1s_err = chron.createVariable('c14_1s_err', 'f4', ('depth_top',))
        c14_1s_err.long_name = '14C 1-sigma error'
        c14_date.units = 'RC yr BP'
        c14_1s_err.missing_value = np.nan
        c14_date[:] = self.chronology_information.df['14C_1s_err'].values

        delta_r = chron.createVariable('delta_r', 'f4', ('depth_top',))
        delta_r.long_name = 'delta R'
        delta_r.missing_value = np.nan
        delta_r[:] = self.chronology_information.df['delta_R'].values

        if 'delta_R_original' in self.chronology_information.df.columns:
            delta_r_orig = chron.createVariable('delta_r_original', 'f4', ('depth_top',))
            delta_r_orig.long_name = 'Original delta R'
            delta_r_orig.description = 'Carbon reservoir correction (delta R) value(s) given in the orignal proxy site data set.'
            delta_r_orig.missing_value = np.nan
            delta_r_orig[:] = self.chronology_information.df['delta_R_original'].values

        delta_r_1s_error = chron.createVariable('delta_r_1s_error', 'f4', ('depth_top',))
        delta_r_1s_error.missing_value = np.nan
        delta_r_1s_error.long_name = 'delta R 1-sigma error'
        delta_r_1s_error[:] = self.chronology_information.df['delta_R_1s_err'].values

        if 'delta_R_1s_err_original' in self.chronology_information.df.columns:
            delta_r_1s_error_orig = chron.createVariable('delta_r_1s_error_original', 'f4', ('depth_top',))
            delta_r_1s_error_orig.long_name = 'Original delta R 1-sigma error'
            delta_r_1s_error_orig.description = 'Carbon reservoir correction 1-sigma error value(s) given in the orignal proxy site data set.'
            delta_r_1s_error_orig.missing_value = np.nan
            delta_r_1s_error_orig[:] = self.chronology_information.df['delta_R_1s_err_original'].values

        other_date = chron.createVariable('other_date', 'f4', ('depth_top',))
        other_date.missing_value = np.nan
        other_date.long_name = 'Other date'
        other_date[:] = self.chronology_information.df['other_date'].values

        other_1s_err = chron.createVariable('other_1s_err', 'f4', ('depth_top',))
        other_1s_err.long_name = 'Other date 1-sigma error'
        other_1s_err.missing_value = np.nan
        other_1s_err[:] = self.chronology_information.df['other_1s_err'].values

        other_type = chron.createVariable('other_type', 'S1', ('depth_top', 'str_dim'))
        other_type.long_name = 'Other date type'
        other_type.missing_value = 'NA'
        other_type._Encoding = 'ascii'
        other_type[:] = _normalize_to_ascii_array(self.chronology_information.df['other_type'].fillna('NA'))

        # Add depth cutoff value attributes to chronology group if
        # self.chronology_information has `cut_shallow` and `cut_deep` attributes.
        if hasattr(self.chronology_information, 'cut_shallow'):
            chron.cut_shallow = self.chronology_information.cut_shallow
        if hasattr(self.chronology_information, 'cut_deep'):
            chron.cut_deep = self.chronology_information.cut_deep

        return chron

    def _attach_data_ncgroup(self, parent):
        """Create and populate data group"""
        data = parent.createGroup('data')
        data.createDimension('depth', None)
        depth = data.createVariable('depth', 'f4', ('depth',), zlib=True)
        depth.long_name = 'Sample depth'
        depth.positive = 'down'
        depth.axis = 'Z'
        depth[:] = self.data.df['depth'].values

        file_depth_unit = str(self.variables['depth'].units)
        if file_depth_unit == '' or file_depth_unit is None:
            depth.units = 'cm'
        else:
            depth.units = file_depth_unit


        age_original = data.createVariable('age_original', 'f4', ('depth',),
                                           zlib=True)
        age_original.missing_value = np.nan
        age_original.long_name = 'Original age'
        age_original[:] = self.data.df['age'].values

        file_age_unit = str(self.variables['age'].units)
        if file_age_unit == '' or file_age_unit is None:
            age_original.units = 'cal years BP'
        else:
            age_original.units = file_age_unit

        if hasattr(self.data, 'age_ensemble') and hasattr(self.data, 'age_median'):
            data.createDimension('draw', self.data.age_ensemble.shape[1])

            age_median = data.createVariable('age_median', 'f4', ('depth',),
                                             zlib=True)
            age_median.units = 'cal years BP'
            age_median.long_name = 'Median age'
            age_median.missing_value = np.nan
            age_median[:] = self.data.age_median['age_median'].values

            agedraw = data.createVariable('age_ensemble', 'f4', ('depth', 'draw'),
                                          zlib=True)
            agedraw.units = 'cal years BP'
            agedraw.long_name = 'Age ensemble'
            agedraw.missing_value = np.nan
            agedraw[:] = self.data.age_ensemble.values

        for col in list(self.data.df.columns):
            col_name = col.lower()

            if col_name in ['depth', 'age']:
                continue

            var = data.createVariable(col_name, 'f4', ('depth',), zlib=True)
            var.missing_value = np.nan

            # Add more attributes to variable.
            attrib_dict = self._variable_attributes(col_name)
            for k, v in attrib_dict.items():
                setattr(var, k, v)

            # Overwrite units attributes with whatever units are given in the NcdcRecord.
            var.units = str(self.variables[col].units)
            var.comments = str(self.variables[col].detail)

            # Grab Mg/Ca cleaning information from "Data Collection Information - Notes"
            if 'mgca' in col_name:
                cleaning_note = str(self.data_collection.notes)
                if 'mg_bcp' in cleaning_note:
                    var.mgca_cleaning_protocol = 'Barker cleaning with hydrogen peroxide'
                elif 'mg_red' in cleaning_note:
                    var.mgca_cleaning_protocol = 'Fully reductive cleaning'
                else:
                    var.mgca_cleaning_protocol = 'NA'
            var[:] = self.data.df[col].values
        return data

    def _attach_ncgroups(self, fl):
        """Dump contents into netCDF4.Dataset

        This is run whenever self.to_netcdf() is called.

        This runs all of the self._attach_*ncgroup() methods and attaches them
        to a netcdf4.Dataset object.
        """
        site_group = self._attach_site_ncgroup(fl)
        # Create and populate chronology group, if chronology_information exists
        if not self.chronology_information.df.empty:
            self._attach_chronology_ncgroup(site_group)
        self._attach_data_ncgroup(site_group)

    def to_netcdf(self, path_or_buffer):
        """Write NcdcRecord contents to a netCDF file
        """
        if isinstance(path_or_buffer, str):
            # Append to file, if it exists, if doesn't exist, create file.
            try:
                with netCDF4.Dataset(filename=path_or_buffer, mode='a', format='NETCDF4') as fl:
                    self._attach_ncgroups(fl)
            except FileNotFoundError:
                with netCDF4.Dataset(filename=path_or_buffer, mode='w', format='NETCDF4') as fl:
                    self._attach_ncgroups(fl)

        else:
            self._attach_ncgroups(path_or_buffer)


class QcPlotMixin:
    """Mixins to add QC plot methods LGM proxy records"""

    @staticmethod
    def _ax_setup(*args, **kwargs):
        try:
            import matplotlib.pylab as plt
        except ModuleNotFoundError:
            raise ModuleNotFoundError('matplotlib needs to be installed for plots')

        return plt.gca(*args, **kwargs)

    def plot_datavariable(self, variable, ax=None):
        """
        Plot a variable in the record data as timeseries.

        The plot also compares the variables from a redated age model with
        the original age model, if the record has been redated.

        Parameters
        ----------
        variable : str
            Name of variable to plot. Must be in ``self.data``.
        ax : :class:`mpl.axes.Axes` or None, optional
            Existing axes to plot onto.

        Returns
        -------
        ax : :class:`mpl.axes.Axes`

        """
        if ax is None:
            ax = self._ax_setup()

        if variable not in self.data.df.columns:
            raise KeyError('{} not found'.format(variable))

        if hasattr(self.data, 'age_median'):
            proxy_df = (self.data.df.set_index('depth')
                        .join(self.data.age_median, lsuffix='__', sort=True))

            new_age = proxy_df.loc[:, ('age_median', variable)].dropna()
            ax.plot(new_age.loc[:, 'age_median'], new_age.loc[:, variable], '.',
                    color='C3', label='MCMC median age')
            ax.plot(new_age.loc[:, 'age_median'], new_age.loc[:, variable],
                    color='C3', linewidth=0.5, label='_nolegend_')
            log.debug('Found new agemodel proxy timeseries')
        else:
            proxy_df = self.data.df.set_index('depth')
            log.debug('Assuming no new agemodel proxy timeseries')

        old_age = proxy_df.loc[:, ('age', variable)].dropna()
        ax.plot(old_age.loc[:, 'age'], old_age.loc[:, variable], 'x',
                color='C0', label='File age')
        ax.plot(old_age.loc[:, 'age'], old_age.loc[:, variable],
                color='C0', linewidth=0.5, label='_nolegend_')

        if 'd18o' in variable.lower():
            ax.invert_yaxis()
        ax.set_ylabel(variable)
        ax.grid()

        return ax

    def plot_sitemap(self, ax=None):
        """
        Plot sample site map using cartopy.

        Requires ``cartopy`` to be installed. Uses the site's northmost latitude
        and easternmost longitude to plot. So, we assume
        ``self.site_information.northernmost_latitude`` and
        ``self.site_information.easternmost_longitude`` are populated.

        Parameters
        ----------
        ax : :class:`mpl.axes.Axes` or None, optional
            Existing axes to plot onto.

        Returns
        -------
        ax : :class:`mpl.axes.Axes`

        """
        try:
            import cartopy.crs as ccrs
            import cartopy.feature as cfeature
        except ModuleNotFoundError:
            raise ModuleNotFoundError('cartopy needs to be installed for mapping')

        latlon = (self.site_information.northernmost_latitude,
                  self.site_information.easternmost_longitude)

        if ax is None:
            ax = self._ax_setup(projection=ccrs.Robinson(central_longitude=latlon[1]))

        ax.set_global()
        ax.add_feature(cfeature.LAND, facecolor='#B0B0B0')
        ax.outline_patch.set_linewidth(0.5)
        ax.plot(latlon[1], latlon[0], 'o', color='C0', transform=ccrs.Geodetic())

        return ax

    def plot_sedrate(self, ax=None):
        """
        Plot prior and posterior sediment rates for record agemodel.

        Requires ``self.chronology_information.bacon_agemodel` to be populated
        with a ``snakebacon.AgeDepthModel``-like instance. You can do this with
        :method:`self.redateredated()`, for example.

        Parameters
        ----------
        ax : :class:`mpl.axes.Axes` or None, optional
            Existing axes to plot onto.

        Returns
        -------
        ax : :class:`mpl.axes.Axes`

        """
        if ax is None:
            ax = self._ax_setup()

        ax = self.chronology_information.bacon_agemodel.plot_sediment_rate(ax)
        ax.lines[-1].set_color('C3')

        return ax

    def plot_sedmemory(self, ax=None):
        """
        Plot prior and posterior sediment memory for record agemodel.

        Requires ``self.chronology_information.bacon_agemodel` to be populated
        with a ``snakebacon.AgeDepthModel``-like instance. You can do this with
        :method:`self.redateredated()`, for example.

        Parameters
        ----------
        ax : :class:`mpl.axes.Axes` or None, optional
            Existing axes to plot onto.

        Returns
        -------
        ax : :class:`mpl.axes.Axes`

        """
        if ax is None:
            ax = self._ax_setup()

        ax = self.chronology_information.bacon_agemodel.plot_sediment_memory(ax)
        ax.lines[-1].set_color('C3')

        return ax

    def plot_agedepth(self, maxage=None, prior_dwidth=30, ax=None):
        """
        Plot age models in relation to core depth.

        Parameters
        ----------
        maxage : float, int, or None, optional
            Cutoff age for the plot age model.
        prior_dwidth : int, optional
            Passed to :method:`snakebacon.AgeDepthModel.plot_prior_dates`.
        ax : :class:`mpl.axes.Axes` or None, optional
            Existing axes to plot onto.

        Returns
        -------
        ax : :class:`mpl.axes.Axes`

        """
        if ax is None:
            ax = self._ax_setup()

        agemodel = self.chronology_information.bacon_agemodel
        data_df = self.data.df
        # Hack to copy and crop the age model to a certain age.
        if maxage is not None:
            agemodel = deepcopy(agemodel)
            too_old = ~(agemodel.age_median() > maxage)
            agemodel._depth = agemodel.depth[too_old]
            agemodel._age_ensemble = agemodel.age_ensemble[too_old]

            # data_df = self.data.df.copy()
            # data_df = data_df.loc[data_df['age'] <= maxage, ('depth', 'age')]

        if len(agemodel._depth) > 0:
            # Skip age model plotting if maxage cut-out all samples.
            ax = agemodel.plot(ax=ax)
            ax.collections[-1].set_cmap('Greys')

            for l in ax.lines:
                l.set_color('C3')

            ax.plot(data_df.loc[:, 'depth'], data_df.loc[:, 'age'], 'C0',
                    label='File age model')
            ax = agemodel.plot_prior_dates(dwidth=prior_dwidth, ax=ax)
            ax.collections[-1].set_color('k')
            ax.collections[-1].set_zorder(10)

            ax.autoscale_view()
        else:
            log.warning('No data age-depth data to plot')

        # Cut cut-off information, if available.
        if hasattr(self.chronology_information, 'cut_shallow'):
            ax.axvline(x=float(self.chronology_information.cut_shallow), color='C4',
                       linestyle='-.', zorder=1.5)
        if hasattr(self.chronology_information, 'cut_deep'):
            ax.axvline(x=float(self.chronology_information.cut_deep), color='C4',
                       linestyle='-.', zorder=1.5)

        ax.set_title('Age model')

        return ax

    def plot_deltar(self, ax=None):
        """
        Plot a description of the site carbon reservoir information.

        Parameters
        ----------
        ax : :class:`mpl.axes.Axes` or None, optional
            Existing axes to plot onto.

        Returns
        -------
        ax : :class:`mpl.axes.Axes`

        """
        if ax is None:
            ax = self._ax_setup()

        latlon = (self.site_information.northernmost_latitude,
                  self.site_information.easternmost_longitude)

        elevation = self.site_information.elevation

        # These checks are not well written.
        try:
            deltar_original = np.int(np.round(self.chronology_information.df['delta_R_original'].values[0]))
            if np.isnan(deltar_original):
                deltar_original = None
        except (KeyError, ValueError, IndexError) as e:
            deltar_original = None

        try:
            deltar_std_original = np.int(np.round(self.chronology_information.df['delta_R_1s_err_original'].values[0]))
            if np.isnan(deltar_std_original):
                deltar_std_original = None
        except (KeyError, ValueError, IndexError) as e:
            deltar_std_original = None

        try:
            deltar_used = np.int(np.round(self.chronology_information.df['delta_R'].values[0]))
        except (KeyError, TypeError):
            deltar_used = None

        try:
            deltar_error_used = np.int(np.round(self.chronology_information.df['delta_R_1s_err'].values[0]))
        except (KeyError, TypeError):
            deltar_error_used = None

        text_template = 'Latitude: {}°\nLongitude: {}°\nElevation: {} m ' \
                        '\n\nΔR: {}\nΔRσ: {}\nFile ΔR: {}\nFile ΔRσ: {}'
        text_str = text_template.format(latlon[0], latlon[1], elevation,
                                        deltar_used, deltar_error_used,
                                        deltar_original, deltar_std_original)

        ax.text(0.05, 0.9, text_str, verticalalignment='top',
                horizontalalignment='left', transform=ax.transAxes)
        ax.set_title('{}\n{}'.format(self.site_information.site_name, datetime.date.today().isoformat()))
        ax.axis('off')

        return ax

    def to_qcpdf(self, pdfpath, proxy_vars=None, plot_agedepth_kws=None):
        """
        Write quality-control report PDF.

        Requires :module:`matplotlib` and :module:`cartopy` to be installed.

        Parameters
        ----------
        pdfpath : str
            Path to write PDF to.
        proxy_vars : iterable or None, optional
            Name of series to include in time series plot. Attempts to use all
            available proxies if ``None``.
        plot_agedepth_kws : dict or None, optional
            Key-word arguments to pass to both call to ``self.plot_agedepth``
            when plotting.
        """
        log.debug('Writing QC report plots')

        try:
            from matplotlib.backends.backend_pdf import PdfPages
            import matplotlib.pylab as plt
        except ModuleNotFoundError:
            raise ModuleNotFoundError('matplotlib needs to be installed for plots')

        try:
            import cartopy.crs as ccrs
        except ModuleNotFoundError:
            raise ModuleNotFoundError('cartopy needs to be installed for mapping')

        if plot_agedepth_kws is None:
            plot_agedepth_kws = {}

        # Find "non-dimension" data variables to plot, if none were passed.
        not_proxy = ['age', 'age_median', 'depth', 'age_ensemble']
        if proxy_vars is None:
            # TODO(brews): This logic might be good candidate for a more general method.
            proxy_vars = [str(x) for x in self.data.df.columns if str(x).lower() not in not_proxy]

        latlon = (self.site_information.northernmost_latitude,
                  self.site_information.easternmost_longitude)

        n_vars = len(proxy_vars)
        n_cols = n_vars + 1

        has_baconagemodel = hasattr(self.chronology_information, 'bacon_agemodel')

        with PdfPages(pdfpath) as pdf:
            fig = plt.figure(figsize=(6.5, 9))

            ax2 = plt.subplot2grid((n_cols, 2), (0, 1),
                                   projection=ccrs.Robinson(central_longitude=latlon[1]))
            ax3 = plt.subplot2grid((n_cols, 2), (0, 0))

            self.plot_sitemap(ax=ax2)

            self.plot_deltar(ax=ax3)

            for i, varstr in enumerate(proxy_vars):
                this_ax = plt.subplot2grid((n_cols, 1), (1 + i, 0), colspan=2)
                this_ax = self.plot_datavariable(varstr, ax=this_ax)
                this_ax.xaxis.label.set_visible(False)
                this_ax.title.set_visible(False)
                if i == 0:
                    this_ax.title.set_visible(True)
                    this_ax.set_title('Proxy variables')

            this_ax.xaxis.label.set_visible(True)
            this_ax.set_xlabel('Age (cal yr BP)')
            fig.tight_layout()
            pdf.savefig(bbox_inches='tight')
            plt.close()

            fig = plt.figure(figsize=(6.5, 9))
            ax1 = plt.subplot2grid((3, 2), (0, 0))
            ax2 = plt.subplot2grid((3, 2), (0, 1))
            ax4 = plt.subplot2grid((3, 2), (1, 0), rowspan=2, colspan=2)

            if has_baconagemodel:
                self.plot_sedrate(ax=ax1)
                self.plot_sedmemory(ax=ax2)
                self.plot_agedepth(maxage=50000, ax=ax4, **plot_agedepth_kws)

            fig.tight_layout()
            pdf.savefig(bbox_inches='tight')
            plt.close()

            fig = plt.figure(figsize=(6.5, 9))
            ax1 = plt.subplot(1, 1, 1)
            if has_baconagemodel:
                self.plot_agedepth(maxage=25000, ax=ax1, **plot_agedepth_kws)
            fig.tight_layout()
            pdf.savefig(bbox_inches='tight')
            plt.close()

            log.debug('QC report plot saved to {}'.format(pdfpath))



import logging
import collections
import datetime
from io import BytesIO
import pandas as pd
from chardet import detect as chdetect
import proxysiphon.records as records


DIVIDER = '#------------------------'
DATALINE_TRIGGER = '# Data lines follow (have no #)'
HEADINGS = ['Contribution_Date', 'Title', 'Investigators',
            'NOTE: Please cite original publication, online resource and date accessed when using this data.',
            'Description and Notes', 'Publication', 'Funding_Agency', 
            'Site Information', 'Data_Collection', 'Species', 
            'Chronology_Information', 'Variables', 'Data']
CHRON_HEADER = '# Labcode\tdepth_top\tdepth_bottom\tmat_dated\t14C_date\t14C_1s_err\tdelta_R\tdelta_R_1s_err\tother_date\tother_1s_err\tother_type\t'
MISSINGVALUE_LABEL = '# Missing Value: '


log = logging.getLogger(__name__)


def grab_latlon(g):
    """Grab latitude from proxychimp.Guts"""
    victim = g.pull_section('Site Information')[0]
    fields = dict(Northernmost_Latitude=None,
                  Southernmost_Latitude=None,
                  Easternmost_Longitude=None,
                  Westernmost_Longitude=None)
    for ln in victim:
        if any([x in ln for x in fields.keys()]):
            ln_split = ln.split(':')
            v = float(ln_split[-1])
            k = ln_split[0][1:].lstrip().rstrip()
            assert k in fields.keys()
            fields[k] = v
    return fields['Northernmost_Latitude'], fields['Westernmost_Longitude']

def grab_elevation(g):
    """Grab site elevation from proxychimp.Guts"""
    victim = g.pull_section('Site Information')[0]
    elevation = None
    for ln in victim:
        if 'Elevation:' in ln:
            ln_split = ln.split(':')
            elevation = float(ln_split[-1])
    return elevation

def grab_collection_year(g):
    """Get collection year from Data_Collection section of proxychimp.Guts"""
    victim = g.pull_section('Data_Collection')[0]
    yr = None
    for ln in victim:
        if 'Collection_Year' in ln:
            ln_split = ln.split(':')
            yr = int(ln_split[-1])
    return yr

def grab_publication_year(g):
    """Get publication year from proxychimp.Guts"""
    victim = g.pull_section('Publication')
    yr = []
    for p in victim:
        for ln in p:
            if 'Published_Date_or_Year' in ln:
                ln_split = ln.split(':')
                yr.append(int(ln_split[-1]))
    return yr

def grab_contribution_date(g):
    """Get contribution date from proxychimp.Guts"""
    victim = g.pull_section('Contribution_Date')[0]
    d = None
    for ln in victim:
        if 'Date' in ln:
            ln_split = ln.split(':')
            d_str = ln_split[-1].split('-')
            assert len(d_str) == 3
            d = datetime.date(int(d_str[0]), int(d_str[1]), int(d_str[2]))
    return d


def find_values(x, k, sep=':', fun=None):
    """Find values in a list of strings containing [k][sep][values]

    Parameters
    ----------
    x : iterable
        Iterable of strings representing lines in a file.
    k : str
        A key or pattern that to search x for.
    sep : str
        A separator that divides the `k` pattern from the target value.
    fun : function-like, optional
        Function used to format target value before returning.

    Returns
    -------
    The target value is returned after removing excess whitespace from left
    and right of the value. If the key or target pattern is not found,
    then None is returned. If an empty string, or whitespace is the found
    value, then None is returned.
    """
    out = None
    for l in x:
        if k in l and sep in l:
            val = l.split(sep, maxsplit=1)[1:][0].lstrip().rstrip()
            if val != '':
                out = val
    if fun is not None and out is not None:
        out = fun(out)
    return out


class Guts:
    """Ghetto and error-tolerant way to parse sections of NCDC proxy text files
    """

    def __init__(self, str_or_path):
        self._section_headings = ['# ' + s for s in HEADINGS]

        self.path = None
        self.encodingguess = None
        self.filestr = None
        try:
            with open(str_or_path, 'rb') as fl:
                flbytes = fl.read()
            self.path = str(str_or_path)
            self.encodingguess = chdetect(flbytes)
            # TODO(brews): Not sure we need ~both~ filestr, and lines + self.data + self.description.
            self.filestr = flbytes.decode(self.encodingguess['encoding'])
        except (FileNotFoundError, OSError):
            self.filestr = str(str_or_path)
        lines = self.filestr.splitlines()

        self.data = []
        self.description = []
        self.data_beginline = None
        self._divide_portions(lines)

        self.sectionindex = None
        self._write_sectionindex()

    def _divide_portions(self, lines):
        """Divide guts into 'description' and 'data' sections
        """
        prev_description = True
        dataline_flag = False
        for n, ln in enumerate(lines):
            if ln[0] == '#' or dataline_flag is False:
                self.description.append(ln)
                if DATALINE_TRIGGER in ln:
                    dataline_flag = True
                    log.debug('Found dataline flag on line {0}'.format(n))
            elif dataline_flag:
                self.data.append(ln)
                if prev_description is True:
                    self.data_beginline = n
                    log.debug('Data portion begins on line {0}'.format(n))
                prev_description = False

    def _write_sectionindex(self):
        """Populate the section index"""
        all_keys = []
        all_start = []
        all_stop = []
        prev_divider = False

        for idx, ln in enumerate(self.description):  # Lines after 6-line title.
            if prev_divider is True and any([h == ln.rstrip() for h in self._section_headings]):
                key = ln[1:].rstrip().lstrip()

                # If already have start idx for other section, append end idx
                # for that section
                if len(all_start) > 0:
                    all_stop.append(idx - 1)

                all_keys.append(key)
                all_start.append(idx)
                prev_divider = False
            if DIVIDER in ln:
                prev_divider = True
        all_stop.append(self.data_beginline)

        section_map = collections.defaultdict(list)
        for (k, start, stop) in zip(all_keys, all_start, all_stop):
            section_map[k].append((start, stop))
        section_map.default_factory = None
        self.sectionindex = section_map

    def pull_section(self, section):
        """Grab a list of list of line strings from the file description for each 'section'"""
        try:
            out = [self.description[slice(*idx)] for idx in self.sectionindex[section]]
        except KeyError:
            raise KeyError('section key "{}" not found'.format(section))
        return out

    def available_sections(self):
        """Get a list of available sections in the file"""
        return list(self.sectionindex.keys())

    def yank_data_df(self):
        """Get 'data' information as dataframe"""
        lines = [x.rstrip() for x in self.data]
        data_bytes = '\n'.join(lines).encode('utf-8')
        missingvalues = self.guess_missingvalues()
        df = pd.read_csv(BytesIO(data_bytes), sep='\t', na_values=missingvalues)
        return df

    def yank_chron_df(self, section_name='Chronology_Information', missingvalues=None):
        """Get chronology information as pandas.DataFrame"""
        if missingvalues is None:
            missingvalues = [-999, 'NaN']
        section = self.pull_section(section_name)[0]
        start_idx = section.index(CHRON_HEADER)
        g_chrond = section[start_idx:]
        g_chrond_cleaned = [x[2:].rstrip() for x in g_chrond]  # Removes the '# ' and ending white space.
        data_bytes = '\n'.join(g_chrond_cleaned).encode('utf-8')
        df = pd.read_csv(BytesIO(data_bytes), sep='\t', na_values=missingvalues)
        return df

    def guess_missingvalues(self):
        """Guess data section missing values"""
        section = self.pull_section('Data')[0]
        out = None
        for ln in section:
            if MISSINGVALUE_LABEL in ln:
                out = ln.split(MISSINGVALUE_LABEL)[1]
        log.debug('Guessed missing value(s): {}'.format(out))
        return out

    def has_data(self):
        """Check if has populated data information"""
        try:
            d = self.yank_data_df()
        except KeyError:
            return False
        if len(d) > 0:
            result = True
        else:
            result = False
        return result

    def has_chron(self):
        """Check if has populated chronology information"""
        try:
            chron = self.yank_chron_df()
        except KeyError:
            return False
        if len(chron) > 0:
            result = True
        else:
            result = False
        return result

    def has_deltar(self):
        """Check if has populated delta R chronology information"""
        try:
            chron = self.yank_chron_df()
        except KeyError:
            return False
        if any(chron.delta_R.notnull()):
            result = True
        else:
            result = False
        return result

    def has_deltar_error(self):
        """Check if has populated delta R error chronology information"""
        try:
            chron = self.yank_chron_df()
        except KeyError:
            return False
        if any(chron.delta_R_1s_err.notnull()):
            result = True
        else:
            result = False
        return result

    def has_datacolumn(self, name):
        """Check if name is in data section columns"""
        try:
            data = self.yank_data_df()
        except KeyError:
            return False
        if name in data.columns:
            result = True
        else:
            result = False
        return result

    def yank_original_source_url(self):
        """Get string of original source URL

        Returns
        -------
        out : str or None

        Raises
        ------
        AssertionError
            If more than one section is found in the file data.
        """
        target_section = 'NOTE: Please cite original publication, online ' \
                         'resource and date accessed when using this data.'
        target_key = 'Original_Source_URL:'

        sections = self.pull_section(target_section)
        assert len(sections) < 2, 'More than one section found'
        section = sections[0]

        try:
            out = find_values(section, target_key, fun=str)
        except AttributeError:
            # target key not found in string
            out = None

        return out

    def yank_data_collection(self):
        """Get data collection information

        Returns
        -------
        out : dict or None

        Raises
        ------
        AssertionError
            If more than one section is found in the file data.
        """
        target_section = 'Data_Collection'
        # List of tuples, tuples give (dict_key, source_key, type_fun)
        target_keys = [('collection_name', 'Collection_Name:', str),
                       ('first_year', 'First_Year:', float),
                       ('last_year', 'Last_Year:', float),
                       ('time_unit', 'Time_Unit:', str),
                       ('core_length', 'Core_Length:', str),
                       ('notes', 'Notes:', str),
                       ('collection_year', 'Collection_Year:', int)]
        out = {k[0]:None for k in target_keys}

        sections = self.pull_section(target_section)
        assert len(sections) < 2, 'More than one section found'
        section = sections[0]

        for dict_key, source_key, type_fun in target_keys:
            out[dict_key] = find_values(section, source_key, fun=type_fun)

        return out

    def yank_description_and_notes(self):
        """Get string of description and notes

        Returns
        -------
        out : str or None

        Raises
        ------
        AssertionError
            If more than one section is found in the file data.
        """
        target_section = 'Description and Notes'
        target_key = 'Description:'

        sections = self.pull_section(target_section)
        assert len(sections) < 2, 'More than one section found'
        section = sections[0]

        try:
            out = find_values(section, target_key, fun=str)
        except AttributeError:
            # target key not found in string
            out = None

        return out

    def yank_publication(self):
        """Get list of publication information

        Returns
        -------
        out : list[dict]
        """
        target_section = 'Publication'
        # List of tuples, tuples give (dict_key, source_key, type_fun)
        target_keys = [('authors', '# Authors:', str),
                       ('published_date_or_year', '# Published_Date_or_Year:', int),
                       ('published_title', '# Published_Title:', str),
                       ('journal_name', '# Journal_Name:', str),
                       ('volume', '# Volume:', str),
                       ('edition', '# Edition:', str),
                       ('issue', '# Issue:', str),
                       ('pages', '# Pages:', str),
                       ('report_number', '# Report Number:', str),
                       ('doi', '# DOI:', str),
                       ('online_resource', '# Online_Resource:', str),
                       ('full_citation', '# Full_Citation:', str),
                       ('abstract', '# Abstract:', str)]
        dict_template = {k[0]:None for k in target_keys}

        out = []
        sections = self.pull_section(target_section)

        for section in sections:

            this_pub = dict_template.copy()
            for dict_key, source_key, type_fun in target_keys:
                this_pub[dict_key] = find_values(section, source_key, fun=type_fun)
            out.append(this_pub)

        return out

    def yank_variables(self):
        """Get variable section information

        Returns
        -------
        out : dict

        Raises
        ------
        AssertionError
            If more than one section is found in the file data.
        """
        target_section = 'Variables'
        sections = self.pull_section(target_section)
        assert len(sections) < 2, 'More than one section found'
        section = sections[0]

        out = {}
        for ln in section:

            # Skip line if not data variables line.
            if ln[:3] != '## ':
                continue

            var_name, components_group = ln[3:].split('\t')
            out[var_name] = tuple(components_group.split(','))

        return out

    def yank_site_information(self):
        """Get site information

        Returns
        -------
        out : dict

        Raises
        ------
        AssertionError
            If more than one section is found in the file data.
        """
        target_section = 'Site Information'
        # List of tuples, tuples give (dict_key, source_key, type_fun)
        target_keys = [('site_name', '# Site_Name:', str),
                       ('location', '# Location:', str),
                       ('country', '# Country:', str),
                       ('northernmost_latitude', '# Northernmost_Latitude:', float),
                       ('southernmost_latitude', '# Southernmost_Latitude:', float),
                       ('easternmost_longitude', '# Easternmost_Longitude:', float),
                       ('westernmost_longitude', '# Westernmost_Longitude:', float),
                       ('elevation', '# Elevation:', float)]
        out = {k[0]:None for k in target_keys}

        sections = self.pull_section(target_section)
        assert len(sections) < 2, 'More than one section found'
        section = sections[0]

        for dict_key, source_key, type_fun in target_keys:
            out[dict_key] = find_values(section, source_key, fun=type_fun)

        return out

    def to_ncdcrecord(self):
        """to NcdcRecord instance"""
        chron = records.ChronologyInformation(df=self.yank_chron_df())
        d = records.Data(df=self.yank_data_df())
        d_collection = records.DataCollection(**self.yank_data_collection())
        description = self.yank_description_and_notes()
        orig_url = self.yank_original_source_url()
        pubs = [records.Publication(**p) for p in self.yank_publication()]
        site_info = records.SiteInformation(**self.yank_site_information())

        vars = {}
        file_vars = self.yank_variables()
        for k,v in file_vars.items():
            vars[k] = records.VariableInfo(*v)

        out = records.NcdcRecord(chronology_information=chron,
                                 data=d,
                                 data_collection=d_collection,
                                 description=description,
                                 original_source_url=orig_url,
                                 publication=pubs,
                                 site_information=site_info,
                                 variables=vars)
        return out



"""Read proxies netCDF5 file and write proxies HDF5 file for LMR Data Assimilation workflow
"""


__all__ = ['nc2lmrh5', 'nc2lmrdf']


import os
import logging
import numpy as np
import pandas as pd
import xarray as xr
import netCDF4
import shapely.affinity
import shapely.geometry

import erebusfall as ef


log = logging.getLogger(__name__)


class DistanceThresholdError(Exception):
    """Raised when the distance between two points is further than a threshold

    Parameters
    ----------
    target_distance : int or float
        The distance between two target points (km).
    distance_threshold : int or float
        The distance threshold.

    """

    def __init__(self, target_distance, distance_threshold):
        self.target_distance = target_distance
        self.distance_threshold = distance_threshold


def chord_distance(latlon1, latlon2):
    """Chordal distance between two sequences of (lat, lon) points

    Parameters
    ----------
    latlon1 : sequence of tuples
        (latitude, longitude) for one set of points.
    latlon2 : sequence of tuples
        A sequence of (latitude, longitude) for another set of points.

    Returns
    -------
    dists : 2d array
        An mxn array of Earth chordal distances [1]_ (km) between points in
        latlon1 and latlon2.

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Chord_(geometry)

    """
    earth_radius = 6378.137  # in km

    latlon1 = np.atleast_2d(latlon1)
    latlon2 = np.atleast_2d(latlon2)

    n = latlon1.shape[0]
    m = latlon2.shape[0]

    paired = np.hstack((np.kron(latlon1, np.ones((m, 1))),
                        np.kron(np.ones((n, 1)), latlon2)))

    latdif = np.deg2rad(paired[:, 0] - paired[:, 2])
    londif = np.deg2rad(paired[:, 1] - paired[:, 3])

    a = np.sin(latdif / 2) ** 2
    b = np.cos(np.deg2rad(paired[:, 0]))
    c = np.cos(np.deg2rad(paired[:, 2]))
    d = np.sin(np.abs(londif) / 2) ** 2

    half_angles = np.arcsin(np.sqrt(a + b * c * d))

    dists = 2 * earth_radius * np.sin(half_angles)

    return dists.reshape(m, n)


def get_nearest(latlon, dain, depth=None, lat_coord='Latitude', lon_coord='Longitude',
                depth_coord='depth', distance_threshold=1500):
    """Get nearest non-NaN to latlon from xarray.DataArray obj

    Finds the nearest not NaN to latlon, and optionally depth. It searches for a
    nearest depth first, if given, and then searches for nearest latlon. Note
    that this does not work with irregular grids, such as rotated polar, etc.

    Parameters
    ----------
    latlon : tuple
        Target latitude and longitude. Must be within -90 to 90 and -180 to 180.
    dain : xarray.DataArray
        Field with regular latlon coordinates.
    depth : float or int, optional
        Target depth to get nearest.
    lat_coord : str, optional
        Name of the latitude coordinate in ``da``.
    lon_coord : str, optional
        Name of the longitude coordinate in ``da``.
    depth_coord : str, optional
        Name of the depth coordinate in ``da``.
    distance_threshold : float or int, optional
        If the nearest distance is larger than this, raise

    Returns
    -------
    nearest : xarray.DataArray
        Nearest points.
    nearest_distance : float
        Chordal distance (km) from target to matched gridpoint.

    Raises
    ------
    DistanceThresholdError
    """
    da = dain.copy()

    assert latlon[0] <= 90 and latlon[0] >= -90
    assert latlon[1] <= 180 and latlon[1] >= -180

    assert lat_coord in da.coords
    assert lon_coord in da.coords

    assert (da[lat_coord].ndim == 1) and (da[lon_coord].ndim == 1)

    # First, find the nearest depth index, if given.
    if depth is not None:
        assert depth_coord in da.coords

        # Note use 'pad' because want next upper water column level value.
        da = da.sortby('depth')
        da = da.sel(**{depth_coord: depth}, method='nearest')

    # Now search for nearest latlon point.
    da_stack = da.stack(yx=[lat_coord, lon_coord]).dropna('yx')
    da_latlon_stack = np.vstack((da_stack[lat_coord], da_stack[lon_coord])).T

    # Any values above 180 become negative -- needed for 0-360 longitudes.
    highlon_msk = da_latlon_stack > 180
    da_latlon_stack[highlon_msk] = da_latlon_stack[highlon_msk] - 360

    distance = chord_distance(np.array([latlon]), da_latlon_stack)
    nearest = da_stack.isel(yx=np.argmin(distance))
    nearest_distance = np.min(distance)

    if nearest_distance > distance_threshold:
        raise DistanceThresholdError(nearest_distance, distance_threshold)

    return nearest, nearest_distance


def get_netcdf_resource(fl, **kwargs):
    """Read NetCDF files as package resource, output for xarray.Dataset

    Parameters
    ----------
    fl : str
        NetCDF resource name.
    **kwargs :
        Passed to ``xarray.open_dataset``.

    Returns
    -------
    data : xarray.Dataset
    """
    here = os.path.abspath(os.path.dirname(__file__))
    flpath = os.path.join(here, fl)
    data = xr.open_dataset(flpath, **kwargs)
    return data


def poly_dateline_wrap(p):
    """Split dateline crossing polygon into multipoly so wraps.

    Parameters
    ----------
    p : shapely.geometry.Polygon

    Returns
    -------
    out : shapely.geometry.Polygon or shapely.geometry.MultiPolygon
    """
    dateline = shapely.geometry.LineString([(180, 90), (180, -90)])

    if not dateline.crosses(p):
        return p

    right_ls = shapely.geometry.LineString([(360, 90), (360, -90)])
    left_ls = shapely.geometry.LineString([(-180, 90), (-180, -90)])
    left_clip = shapely.geometry.MultiLineString([left_ls, dateline]).convex_hull
    right_clip = shapely.geometry.MultiLineString([dateline, right_ls]).convex_hull

    northpacific_left = p.intersection(left_clip)
    dl_right_off = p.intersection(right_clip)

    # Shift right of dateline back so lon (x) is within [-180, 180]
    northpacific_lr = shapely.affinity.translate(shapely.geometry.LinearRing(dl_right_off.exterior.coords),
                                                 xoff=-360)
    northpacific_right = shapely.geometry.Polygon(northpacific_lr)
    out = shapely.geometry.MultiPolygon([northpacific_left, northpacific_right])
    return out


def find_seasonality(sitegrp, proxy_grp):
    """Return string list of ints giving site variable seasonality.

    Parameters
    ----------
    sitegrp : netCDF4.Group
        Proxy site netcdf group.
    proxy_grp: netCDF4.Group
        Proxy variable netcdf group

    Returns
    -------
    out : str
    """
    out = list(range(1, 13))

    proxy_type = str(proxy_grp.long_name)
    if proxy_type == "UK'37":
        log.debug('finding seasonality for UK37 proxy')
        # Jess Tierney polygons from BAYSPLINE
        mediterranean = shapely.geometry.Polygon([(-5.5, 36.25),
                                                  (3, 47.5),
                                                  (45, 47.5),
                                                  (45, 30),
                                                  (-5.5, 30)])
        northatlantic = shapely.geometry.Polygon([(-55, 48),
                                                  (-50, 70),
                                                  (20, 70),
                                                  (10, 62.5),
                                                  (-4.5, 58.2),
                                                  (-4.5, 48)])
        northpacific_raw = shapely.geometry.Polygon([(135, 45),
                                                     (135, 70),
                                                     (250, 70),
                                                     (232, 52.5),
                                                     (180, 45)])
        northpacific = poly_dateline_wrap(northpacific_raw)

        latlon = (float(sitegrp.latitude), float(sitegrp.longitude))
        assert -90 < latlon[0] < 90, 'site latitude must be -90 < lat < 90'
        assert -180 < latlon[1] < 180, 'site longitude must be -180 < lon < 180'

        site_location = shapely.geometry.Point(latlon[::-1])
        if mediterranean.contains(site_location):
            out = [1, 2, 3, 4, 5, 11, 12]
        if northatlantic.contains(site_location):
            out = [8, 9, 10]
        if northpacific.contains(site_location):
            out = [6, 7, 8]

    elif proxy_type == 'd18O':
        foram_type = str(proxy_grp.foraminifera_type)
        latlon = (float(sitegrp.latitude), float(sitegrp.longitude))
        log.debug('finding seasonality for d18O proxy ({}) @ {}'.format(foram_type, latlon))
        foram_seas = get_netcdf_resource('foram_seasons.nc', group='d18oc')
        foram_seas = foram_seas.where(foram_seas != '')  # Replace '' with nan.
        # Need to normalize full species/subspecies names to the variable names used
        # in the FORAM_SEASONS_NETCDF group.
        foraminifera_map = {'Globigerina bulloides': 'G. bulloides',
                            'Neogloboquadrina pachyderma sinistral': 'N. pachyderma',
                            'Neogloboquadrina incompta': 'N. incompta',
                            'Globigerinoides ruber white': 'G. ruber',
                            'Globigerinoides ruber pink': 'G. ruber',
                            'Trilobatus sacculifer': 'T. sacculifer',
                            }
        foram_spp_str = foraminifera_map.get(foram_type)
        if foram_spp_str is not None:
            nearest_seas, _ = get_nearest(latlon, foram_seas[foram_spp_str],
                                          lat_coord='lat', lon_coord='lon')
            out = nearest_seas.item()

        else:
            pass  # foram type not in our species mapping so use annual.

    elif proxy_type == 'Mg/Ca':
        foram_type = str(proxy_grp.foraminifera_type)
        latlon = (float(sitegrp.latitude), float(sitegrp.longitude))
        log.debug('finding seasonality for Mg/Ca proxy ({}) @ {}'.format(foram_type, latlon))
        foram_seas = get_netcdf_resource('foram_seasons.nc', group='mgca')
        foram_seas = foram_seas.where(foram_seas != '')  # Replace '' with nan.
        # Need to normalize full species/subspecies names to the variable names used
        # in the FORAM_SEASONS_NETCDF group.
        foraminifera_map = {'Globigerina bulloides': 'G. bulloides',
                            'Neogloboquadrina pachyderma sinistral': 'N. pachyderma sinistral',
                            'Neogloboquadrina incompta': 'N. pachyderma dextral',
                            'Globigerinoides ruber white': 'G. ruber white',
                            'Globigerinoides ruber pink': 'G. ruber pink',
                            'Trilobatus sacculifer': 'G. sacculifer',
                            }
        foram_spp_str = foraminifera_map.get(foram_type)
        if foram_spp_str is not None:
            nearest_seas, _ = get_nearest(latlon, foram_seas[foram_spp_str],
                                          lat_coord='lat', lon_coord='lon')
            out = nearest_seas.item()
        else:
            pass  # foram type not in our species mapping so use annual.
    log.debug('season months are: {}'.format(out))
    return str(out)


def icevol_correction(proxies, metadata):
    """Apply ice-volume correction to d18O proxies in LMR proxy dataframes.

    Returns modified copies of the original data.
    """
    p = proxies.copy()
    m = metadata.copy()

    matches = ['d18o' in c.lower() for c in p.columns]

    if not any(matches):
        return p, m

    m.set_index('Proxy ID', inplace=True)

    matches_idx = np.where(matches)
    matched_columns = p.columns[matches_idx]

    for c in matched_columns:
        proxy_raw = proxies[c].values
        age_raw = proxies[c].index.values
        age_yr = 1950 - age_raw  # CE/BC to Yr BP
        proxy_adjusted = ef.icevol_correction(age_yr, proxy_raw, proxytype='d18o',
                                              timeunit='ya')
        p.loc[:, c] = proxy_adjusted

        m.loc[c, 'Oldest (C.E.)'] = p[c].dropna().index.min()
        m.loc[c, 'Youngest (C.E.)'] = p[c].dropna().index.max()

    m.reset_index(inplace=True)
    return p, m


def lmr_da_dfs(sitegrp=None, agemodel_iter=None, find_modern_seasonality=True):
    """Return proxy data and metadata pandas df needed for LMR DA proxy input

    If no args are passed, return empty proxy and metadata (template) pandas
    dataframes.

    Parameters
    ----------
    sitegrp : netCDF4.Group or None, optional
        Proxy site netcdf group.
    agemodel_iter : int or None, optional
        Age-depth model iteration to use, if available. This is the index of the
        'agemodel_ensemble' column to use for the output data. If ``None`` (default),
        uses median age.
    find_modern_seasonality : bool
        Do you want to estimate sample seasonality using the modern record?
        Sets annual seasonality if False. Default is True.

    Returns
    -------
    data : pandas.DataFrame
    meta : pandas.DataFrame
    """
    data_template = pd.DataFrame()
    meta_template = pd.DataFrame({'Proxy ID': [], 'Site': [], 'Lat (N)': [], 'Lon (E)': [],
                                  'Archive type': [], 'Proxy measurement': [],
                                  'Resolution (yr)': [], 'Reference': [], 'Databases': [],
                                  'Seasonality': [], 'Elev': [],
                                  'Oldest (C.E.)': [], 'Youngest (C.E.)': []})
    meta_template = meta_template.astype(dtype={'Proxy ID': 'object', 'Site': 'object',
                                                'Lat (N)': 'float64', 'Lon (E)': 'float64',
                                                'Archive type': 'object',
                                                'Proxy measurement': 'object',
                                                'Resolution (yr)': 'float64',
                                                'Reference': 'object', 'Databases': 'object',
                                                'Seasonality': 'object', 'Elev': 'float64',
                                                'Oldest (C.E.)': 'float64',
                                                'Youngest (C.E.)': 'float64'})

    if sitegrp is None:
        return data_template.copy(), meta_template.copy()

    log.info('extracting LMR data from {}'.format(str(sitegrp.site_name)))

    variables_to_skip = ['depth', 'age_original', 'age_median', 'age_ensemble']
    proxy_variables = [(k, v) for k, v in sitegrp['data'].variables.items() if k not in variables_to_skip]

    latitude = float(sitegrp.latitude)
    longitude = float(sitegrp.longitude)
    elevation = float(sitegrp.elevation)
    data_df = data_template.copy()
    meta_df = meta_template.copy()

    for k, v in proxy_variables:
        log.debug('processing variable {}'.format(str(k)))

        # Convert years BP to CE/BP and find min max.
        if agemodel_iter is None:
            try:
                age_yrs_ce = 1950 - sitegrp['data'].variables['age_median'][:]
            except KeyError:
                age_yrs_ce = 1950 - sitegrp['data'].variables['age_original'][:]
        else:
            idx = int(agemodel_iter)
            age_yrs_ce = 1950 - sitegrp['data'].variables['age_ensemble'][:, idx]

        # Make depth cutoff mask from depth cutoffs, if available.
        cut_deep = np.inf
        if hasattr(sitegrp['chronology'], 'cut_deep'):
            cut_deep = float(sitegrp['chronology'].cut_deep)
        cut_shallow = -np.inf
        if hasattr(sitegrp['chronology'], 'cut_shallow'):
            cut_shallow = float(sitegrp['chronology'].cut_shallow)
        depth = sitegrp['data'].variables['depth'][:]
        cutoff_msk = (depth >= cut_shallow) & (depth <= cut_deep)

        notnan_and_notcut = ~np.isnan(v[:].filled(np.nan)) & cutoff_msk
        youngest_ce = np.max(age_yrs_ce[notnan_and_notcut])
        oldest_ce = np.min(age_yrs_ce[notnan_and_notcut])

        # Put together proxy ID and proxy measurement strings.
        siteid = str(sitegrp.site_name).strip().lower()
        pmeasurement = str(k).strip().lower()
        # Append cleaning protocol info if available for Mg/Ca
        try:
            if v.long_name == 'Mg/Ca':
                log.debug('Mg/Ca variable, attempting to find cleaning protocol')
                # Should throw error if have Mg/Ca record w/o cleaning protocol info.
                cleaning_protocol = str(v.mgca_cleaning_protocol)
                cleaning_str = None
                if 'reductive' in cleaning_protocol.lower():
                    cleaning_str = ':reductive'
                elif 'barker' in cleaning_protocol.lower():
                    cleaning_str = ':barker'
                pmeasurement += cleaning_str
        except AttributeError:  # caught if v doesn't have long_name attrib...
            pass
        proxyid = siteid + ':' + pmeasurement

        this_meta = meta_template.copy()
        this_meta['Proxy ID'] = [proxyid]
        this_meta['Site'] = [siteid]
        this_meta['Lat (N)'] = latitude
        this_meta['Lon (E)'] = longitude
        this_meta['Archive type'] = ['Marine sediments']
        this_meta['Proxy measurement'] = [pmeasurement]
        this_meta['Resolution (yr)'] = [(youngest_ce - oldest_ce) / len(age_yrs_ce[notnan_and_notcut])]
        this_meta['Reference'] = [str(None)]
        this_meta['Databases'] = ['[DTDA]']
        if find_modern_seasonality:
            this_meta['Seasonality'] = find_seasonality(sitegrp, v)
        else:
            this_meta['Seasonality'] = str(list(range(1, 13)))
        this_meta['Elev'] = elevation
        this_meta['Oldest (C.E.)'] = oldest_ce
        this_meta['Youngest (C.E.)'] = youngest_ce
        meta_df = meta_df.append(this_meta, ignore_index=True)

        d = (pd.DataFrame({'Year C.E.': age_yrs_ce[cutoff_msk],
                           proxyid: v[:].filled(np.nan)[cutoff_msk]})
               .set_index('Year C.E.')
               .dropna(how='any'))
        data_df = data_df.join(d, how='outer')

    data_df = data_df.sort_index(ascending=False)

    return data_df, meta_df


def _lmr_df_from_nc_sites(fl, agemodel_iter=None, find_modern_seasonality=True):
    """Create LMR data and metadata dataframes from opened netCDF file group"""
    all_data_df, all_meta_df = lmr_da_dfs()

    for site_grp in fl.groups.values():
        try:
            site_data_df, site_meta_df = lmr_da_dfs(site_grp,
                                                    agemodel_iter=agemodel_iter,
                                                    find_modern_seasonality=find_modern_seasonality)
        except TypeError as e:
            errormsg = '{} raised - skipping file - {} - Mg/Ca variable likely missing cleaning protocol info'
            log.error(errormsg.format(e, site_grp.site_name))
            continue

        all_meta_df = all_meta_df.append(site_meta_df, ignore_index=True)
        all_data_df = all_data_df.join(site_data_df, how='outer')

    return all_meta_df, all_data_df


def nc2lmrdf(path_or_buffer, agemodel_iter=None,
             icevol_cor=True, find_modern_seasonality=True):
    """Read proxy netCDF and output to LMR DA-format dataframes.

    Parameters
    ----------
    path_or_buffer : str or netCDF4.Dataset
        Input proxy netCDF file path or opened buffer.
    agemodel_iter : int or None, optional
        Optional index of age-model iteration to use in output -- if multiple
        age-model iterations are available.
    icevol_cor : bool, optional
        Do you want to apply ice-volume correction to the d18O foraminiferal
        records? This is done with the `erebusfalls` package. `True` by default.
    find_modern_seasonality : bool, optional
        Do you want to estimate sample seasonality using the modern record?
        Sets annual seasonality if False. Default is True.

    Returns
    -------
    all_data_df : pd.Dataframe
        LMR proxy data.
    all_meta_df : pd.Dataframe
        LMR proxy metadata.

    Also see
    --------
    `proxysiphon.nc2lmrh5`

    """
    # Open proxy site netCDF file and collect needed data or read directly from
    # obj if it's already open.
    if isinstance(path_or_buffer, str):
        with netCDF4.Dataset(filename=path_or_buffer, mode='r') as fl:
            all_meta_df, all_data_df = _lmr_df_from_nc_sites(fl,
                                                             agemodel_iter=agemodel_iter,
                                                             find_modern_seasonality=find_modern_seasonality)
    else:
        all_meta_df, all_data_df = _lmr_df_from_nc_sites(path_or_buffer,
                                                         agemodel_iter=agemodel_iter,
                                                         find_modern_seasonality=find_modern_seasonality)

    all_data_df = all_data_df.sort_index(ascending=False)

    if icevol_cor:
        # Ice-volume correction to d18O foram proxies:
        all_data_df, all_meta_df = icevol_correction(all_data_df, all_meta_df)

    return all_data_df, all_meta_df


def nc2lmrh5(path_or_buffer, h5file, agemodel_iter=None, icevol_cor=True,
             find_modern_seasonality=True):
    """Read proxy netCDF and output to LMR DA-format HDF5 file.

    Parameters
    ----------
    path_or_buffer : str or netCDF4.Dataset
        Input proxy netCDF file path or opened buffer.
    h5file : str
        Path to output HDF5 file. Not written if None.
    agemodel_iter : int or None, optional
        Optional index of age-model iteration to use in output -- if multiple
        age-model iterations are available.
    icevol_cor : bool, optional
        Do you want to apply ice-volume correction to the d18O foraminiferal
        records? This is done with the `erebusfalls` package. `True` by default.
    find_modern_seasonality : bool, optional
        Do you want to estimate sample seasonality using the modern record?
        Sets annual seasonality if False. Default is True.

    Returns
    -------
    Nothing is returned, output is written to `h5file`.

    Also see
    --------
    `proxysiphon.nc2lmrdf`
    """
    all_data_df, all_meta_df = nc2lmrdf(path_or_buffer, agemodel_iter=agemodel_iter,
                                        icevol_cor=icevol_cor,
                                        find_modern_seasonality=find_modern_seasonality)

    # Write to H5 file.
    log.debug('writing to HDF5 file: {}'.format(h5file))
    all_meta_df.to_hdf(h5file, key='meta', mode='w', format='table',
                       complevel=9, complib='blosc')
    all_data_df.to_hdf(h5file, key='proxy', mode='r+', format='table',
                       complevel=9, complib='blosc')
